// -*- mode: c++; mode: fold -*-
// Description								/*{{{*/
// $Id: dudf.cc,v 1.2 2002/07/25 18:07:18 prrt Exp $
/* ######################################################################
   
    DUDF - Distribution Upgradeability Description Format
   
   Implementation of DUDF
   It is responsible for generating the DUDF file as defined in MANCOOSI 
   project.
   DUDF contais one problem instance.
   Inspired on GlobalError class written by Jason Gunthorpe.
   
   ##################################################################### */
									/*}}}*/
// Include Files							/*{{{*/
#ifdef __GNUG__
#pragma implementation "apt-pkg/dudf.h"
#endif 


#include <apt-pkg/dudf.h>
#include <apt-pkg/error.h>
#include <apt-pkg/fileutl.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/md5.h>
#include <apt-pkg/cmndline.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <iostream>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>

#include <stdarg.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <locale.h>


#include <config.h>

#include <apti18n.h>
#include <fstream>
#include <iostream>

#include <apt-pkg/configuration.h>
   									/*}}}*/

using namespace std;

// Dudf Object							/*{{{*/

 GlobalDudf *_GetDudfObj()
 {
    static GlobalDudf *Obj = new GlobalDudf;
    return Obj;
 }

									/*}}}*/

// GlobalDudf::GlobalDudf - Constructor				/*{{{*/
// ---------------------------------------------------------------------
/* */
GlobalDudf::GlobalDudf() 
{
	time_t Date;
	struct tm * timeinfo;
	char buffer [80];

	setlocale(LC_TIME,"en_US");
	
	time(&Date);
	timeinfo = localtime ( &Date );
	
	strftime (buffer,80,"%Y%m%d%H%M%S",timeinfo);   
	
	DudfErrorMsg="";
	FilenameFingerprint=buffer;

	strftime (buffer,80,"%a, %d %b %Y %H:%M:%S GMT",timeinfo);   
	Timestamp=buffer;
	
	PtrPkgInfo= new vector<DudfPackageInfo>();

}

bool GlobalDudf::generateUUID(string &uuid)
{

    FILE *rpmqstream;
    int UUIDSIZE=37; // 36 chars for uuid + \0
    char tmpbuffer[UUIDSIZE]; 

    rpmqstream = popen("/usr/bin/uuidgen -t","r");

    if (rpmqstream==NULL)
	return _error->Errno("generateUUID()",_("Unable to popen() uuidgen"));

    fgets(tmpbuffer, UUIDSIZE, rpmqstream);

    uuid = string(tmpbuffer);
    return true;

}
									/*}}}*/


//  GlobalDudf::DumpToFile- Dump DUDF into XML file		/*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::DumpToFile(CommandLine &CmdL, string ResultStatus)
{

    string DudfDirName = _config->Find("Dir::Dudf");
    if (DudfDirName == "")
	DudfDirName="/var/lib/apt/dudf";

    string dudfFilename =  DudfDirName + "/dudf_" + FilenameFingerprint  + ".xml";
    string pkgstatus;

    string UUID;
    generateUUID(UUID);
    xmlDocPtr doc = NULL;       
    xmlNodePtr root_node = NULL, node = NULL, node1 = NULL , node2 = NULL , node3 = NULL  ;
    xmlDtdPtr dtd = NULL;       
    xmlNsPtr  ns = NULL;
    char buff[256];
    int i, j;

   
    doc = xmlNewDoc(BAD_CAST "1.0");
    root_node = xmlNewNode(ns , BAD_CAST "dudf");
    ns = xmlNewNs ( root_node, BAD_CAST "http://www.mancoosi.org/2008/cudf/dudf", BAD_CAST "dudf");
    xmlDocSetRootElement(doc, root_node);

    /* 
     * Timestamp
     */
    /* Time is the same used for fingerprinting files */ 
    node1 = xmlNewChild(root_node, NULL, BAD_CAST "timestamp", BAD_CAST Timestamp.c_str());

    /*
     * UID
     */
    node1 = xmlNewChild(root_node, NULL, BAD_CAST "uid", BAD_CAST UUID.c_str());


    /* 
     * The Distribution node and content
     */
    
    SetDistro();
   
    xmlNewChild(root_node, NULL, BAD_CAST "distribution",
                BAD_CAST DudfInfoElement.Distribution.c_str());

    /* 
     * The installer  node. 
     */
    node = xmlNewChild(root_node, NULL, BAD_CAST "installer", NULL);
    node1 = xmlNewChild(node, NULL, BAD_CAST "name", BAD_CAST "rpm"); 
    node1 = xmlNewChild(node, NULL, BAD_CAST "version", BAD_CAST RPM_VERSION_RAW );
    
    /* 
     * The meta-installer  node
     */
    node = xmlNewChild(root_node, NULL, BAD_CAST "meta-installer", NULL);
    node1 = xmlNewChild(node, NULL, BAD_CAST "name", BAD_CAST PACKAGE); 
    node1 = xmlNewChild(node, NULL, BAD_CAST "version", BAD_CAST VERSION);
    

    /* 
     * The problem node
     */
    node = xmlNewChild(root_node, NULL, BAD_CAST "problem", NULL);

// PT: /usr/lib/rpm/rpmq -q --all -i --qf [%{REQUIRENAME}%{REQUIREFLAGS:depflags}%{REQUIREVERSION}\\n] --qf [PROV%{PROVIDENAME}%\|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags}%{PROVIDEVERSION}}:{}\|\n]
//  /usr/lib/rpm/rpmq -qa  --qf [%{NAME}\ -\ ]  --qf [%{EPOCH}:]   --qf [%{VERSION}\ \ -%{RELEASE}\\n] 

    
    if(ResultStatus != "success"){

    	/*  The package status                              */ 
    	node1 = xmlNewChild(node, NULL, BAD_CAST "package-status", NULL);
    	node2 = xmlNewChild(node1, NULL, BAD_CAST "installer", NULL);

	if ( PackageDb(pkgstatus) != true )
	    _error->Error(_("Can not retrieve installer status."));

	node3 = xmlNewChild(node1, NULL, BAD_CAST "status", BAD_CAST pkgstatus.c_str());
//	free (pkgstatus);

    	string PkgDbFilename=_config->Find("Dir::State::status","/var/lib/rpm/Packages").c_str();
	
	xmlNewProp(node3, BAD_CAST "dudf:filename", BAD_CAST PkgDbFilename.c_str());

    
    }

    vector<string> *PtrListFiles= new vector<string>();    
    if((ResultStatus != "success") && PackageUniverse(PtrListFiles)) {
    	/*  The package Universe*/ 
    	node1 = xmlNewChild(node, NULL, BAD_CAST "package-universe", NULL);
    	string ListsFilename= DudfDirName + "/lists_" + FilenameFingerprint  + ".tar.bz2 ";
    	xmlNewProp(node1, BAD_CAST "filename", BAD_CAST ListsFilename.c_str());

    	for (vector<string>::iterator I = PtrListFiles->begin(); I != PtrListFiles->end(); I++) {
    		
    		node2 = xmlNewChild(node1, NULL, BAD_CAST "package-list", NULL);
    		
    		MD5Summation md5sum = MD5Summation();
    		    		

		if (strstr(I->c_str(), "synthesis.hdlist") != NULL)
    			xmlNewProp(node2, BAD_CAST "dudf:format", BAD_CAST "synthesis_hdlist" );
		else if (strstr(I->c_str(), "hdlist") != NULL)
    			xmlNewProp(node2, BAD_CAST "dudf:format", BAD_CAST "hdlist" );
    		xmlNewProp(node2, BAD_CAST "dudf:filename", BAD_CAST I->c_str());
    		
    		/* Not available. Need to parse source lists */ 
   		    /* xmlNewProp(node2, BAD_CAST "format", BAD_CAST ListsFilename.c_str()); */
    		
    		string NameFileFd="/" + _config->FindDir("Dir::State::lists") + I->c_str(); 
    		FileFd F(NameFileFd, FileFd::ReadOnly);
    		
    		if (! F.Eof()) {
    		
    			// This MD5 is of the uncompressed file. In the server, the hash should
			// also be generated from the uncompressed file.
			// For the moment, is only working for synthesis since hdlists raw are stored
			// differently locally and in the server. 
			// For hdlists we sould have to use MD5SUM releases file in the server.
    			
    			md5sum.AddFD(F.Fd(), F.Size());
    			node3 = xmlNewChild(node2, NULL, BAD_CAST "hash", BAD_CAST md5sum.Result().Value().c_str());
    		}
    	
    	}    
    	
    }
    delete PtrListFiles;
    
    
    /*  The package universe */ 
    node1 = xmlNewChild(node, NULL, BAD_CAST "action", NULL);
        
    
    string CmdCall="apt-get";
    for (const char **I = CmdL.FileList; *I != 0; I++)
    	 CmdCall= CmdCall +" " + *I;    	
 		 
    
	node2 = xmlNewChild(node1, NULL, BAD_CAST "command", BAD_CAST CmdCall.c_str());
    
	/* 
     * The outcome node
     */
    node = xmlNewChild(root_node, NULL, BAD_CAST "outcome", NULL);

    /*  Packages */ 
    for (vector<DudfPackageInfo>::iterator I = PtrPkgInfo->begin(); I != PtrPkgInfo->end(); I++) {

    	if  (I->Status == "apparently installable") {
    		node1 = xmlNewChild(node, NULL, BAD_CAST "package", BAD_CAST I->Status.c_str());
    		xmlNewProp(node1, BAD_CAST "name", BAD_CAST I->Name.c_str());
    		xmlNewProp(node1, BAD_CAST "version-release", BAD_CAST I->VerRel.c_str());
    		
    	}
    	else  
    	{
    		node1 = xmlNewChild(node, NULL, BAD_CAST "package", NULL);
    		xmlNewProp(node1, BAD_CAST "name", BAD_CAST I->Name.c_str());
    		xmlNewProp(node1, BAD_CAST "version-release", BAD_CAST I->VerRel.c_str());
    		node2 = xmlNewChild(node1, NULL, BAD_CAST I->Status.c_str(), NULL);
    		xmlNewProp(node2, BAD_CAST "name", BAD_CAST BAD_CAST I->DepBrokenName.c_str());
    		xmlNewProp(node2, BAD_CAST "constrain", BAD_CAST BAD_CAST I->DepConstrain.c_str());
    		xmlNewProp(node2, BAD_CAST "version", BAD_CAST BAD_CAST I->DepConstVersion.c_str());
    		node3 = xmlNewChild(node2, NULL, BAD_CAST "package" , NULL);
    		xmlNewProp(node3, BAD_CAST "name", BAD_CAST I->DepBrokenName.c_str());
    		if (! I->DepCandidate.empty())
	    		xmlNewProp(node3, BAD_CAST "candidate", BAD_CAST I->DepCandidate.c_str());
    	}
    	
    
    	
    }
    
    /*  Result */ 
    node1 = xmlNewChild(node, NULL, BAD_CAST "result", BAD_CAST ResultStatus.c_str());

    /*  Error msg - not printed on sucess */
    if (! DudfErrorMsg.empty())
    	node1 = xmlNewChild(node, NULL, BAD_CAST "error", BAD_CAST DudfErrorMsg.c_str());

    

    xmlSaveFormatFileEnc( dudfFilename.c_str() , doc, "UTF-8", 1);

   
    xmlFreeDoc(doc);

   
    xmlCleanupParser();
    
    
    return(0);

	
   return true;
}
									/*}}}*/


//  GlobalDudf::SetDistro - Set distribution information		/*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::SetDistro()
{


    string File = "/etc/release";

    // See if the file exists
    struct stat Buf;
    if (stat(File.c_str(),&Buf) != 0)
        _error->Error(_("File releases does not exist:"), File.c_str());


    // Open the stream for reading
    ifstream F(File.c_str(),ios::in);

    if (!F != 0) 
        _error->Error(_("Could not open releases file:"), File.c_str());


    char Buffer[1024];

    int CurLine = 0;
    while (F.eof() == false)
    {
        F.getline(Buffer,sizeof(Buffer));
        CurLine++;

        if (F.fail() && !F.eof())
        	_error->Error(_("Error reading releases file:"), File.c_str());


        if (strlen(Buffer) == 0)
            break;
        
        DudfInfoElement.Distribution=Buffer;

    }
    

   return true;
}
									/*}}}*/



//  GlobalDudf::PackageDb - Include intensional / extensional Package DB		/*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::PackageDb(string &pkgstatus)
{
	


    // I have to execute this process and store the result in JSON inside the  field
//  /usr/lib/rpm/rpmq -qa --qf [\\[\"%{NAME}\",\"] --qf [%{EPOCH}:]  --qf [%{VERSION}\",\"] --qf [%{RELEASE}\",\\[] --qf [\"d\":\"%{REQUIRENAME}%{REQUIREFLAGS:depflags}%{REQUIREVERSION}\"\,]    --qf [\\],\"%{RPMVERSION}\",\\[]  --qf [\"%{PROVIDENAME}%\|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags}%{PROVIDEVERSION}}:{}\|\",]  --qf [\\],\"%{ARCH}\"]  --qf [\"%{CONFLICTNAME}%\|CONFLICTFLAGS?{%{CONFLICTFLAGS:depflags}%{CONFLICTVERSION}}:{}\|\",]   --qf [\\],\"%{INSTALLTID}\",\\]\\n] 

    /* JSON Definition
       { "package-status" : [ ["name","epoch:version","release",["dep2,"dep2",..],"rpm_version",["prov1","prov2",...] ],[pkg_2],...,[pkg_n]] } */


    FILE *rpmqstream;
    char  tmpbuffer[MAXBUFFER];

    pkgstatus=" { \"package-status\" : [ ";

    rpmqstream = popen("/bin/rpm -qa --qf [\\\\[\\\"%{NAME}\\\",\\\"]  --qf [%{EPOCH}:]  --qf [%{VERSION}\\\",\\\"] --qf [%{RELEASE}\\\",\\\\[] --qf [\\\"%{REQUIRENAME}%{REQUIREFLAGS:depflags}%{REQUIREVERSION}\\\",] --qf [\\\\],\\\"%{RPMVERSION}\\\",\\\\[]  --qf [\\\"%{PROVIDENAME}%\\|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags}%{PROVIDEVERSION}}:{}\\|\\\",]  --qf [\\\\],\\\"%{ARCH}\\\",\\\\[]    --qf [\\\"%{CONFLICTNAME}%\\|CONFLICTFLAGS?{%{CONFLICTFLAGS:depflags}%{CONFLICTVERSION}}:{}\\|\\\",]  --qf [\\\\],\\\"%{INSTALLTID}\\\"\\\\],\\\\n]  ", "r");

    if (rpmqstream==NULL)
	return _error->Errno("rpmq",_("Unable to get the status of the packages in the system"));

    while (fgets(tmpbuffer, MAXBUFFER, rpmqstream))
	pkgstatus+=tmpbuffer;

    string::iterator it = pkgstatus.end()-2; /* That sneaky last comma... */
    pkgstatus.erase(it);

    pkgstatus+="]}";
    
	    
    pclose(rpmqstream);
	/*
	 * agrr:  Hack alert!: We should be using a proper JSON library 
	 */
    string copy=pkgstatus;
    int i;
  	for (i=0; i <copy.length()-2; i++)
  	{
	    if(copy[i]==',' && copy[i+1]==']' && copy[i+2]==',')  /* Wooohooo */
		    copy[i]=' ';
  	}
	pkgstatus=copy;

    return true;
}
/*}}}*/


//  GlobalDudf::PackageUniverse - Include intensional / extensional Package Lists		/*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::PackageUniverse(vector<string> *PtrListFiles)
{
	

	
	string DirLists= "/" + _config->FindDir("Dir::State::lists");
	

	DIR *D = opendir(DirLists.c_str());
	if (D == 0)
		return _error->Errno("opendir",_("Unable to read %s"),DirLists.c_str());

	string StartDir = SafeGetCWD();
	if (chdir(DirLists.c_str()) != 0)
	{
		closedir(D);
		return _error->Errno("chdir",_("Unable to change to %s"),DirLists.c_str());
	}

	for (struct dirent *Dir = readdir(D); Dir != 0; Dir = readdir(D))
	{
		// Skip files that aren't lists
		if (strcmp(Dir->d_name,"lock") == 0 ||
				strcmp(Dir->d_name,"partial") == 0 ||
				strcmp(Dir->d_name,".") == 0 ||
				strcmp(Dir->d_name,"..") == 0)
			continue;
		
		PtrListFiles->push_back(Dir->d_name);
		
		//Args[++argcount] = Dir->d_name;	
		
	};


	/* 
	 * This code was when lists were included in a separate file 

	// Fork tar
	int Process = fork();
	const char *Args[PtrListFiles->size()];
	
	if (Process < 0)
		return _error->Errno("fork","Couldn't fork tar");

	// The child
	if (Process == 0)
	{
		// tar util existence should be ckecked...
		string tarcmd="/bin/tar";
		
		FileFd FdSilentCout("/dev/null" , FileFd::WriteEmpty);
		dup2(FdSilentCout.Fd(),STDOUT_FILENO);
		SetCloseExec(STDOUT_FILENO,false);
				
		string ListsFilename=_config->Find("Dir::Dudf") + "/lists_" + FilenameFingerprint  + ".tar.bz2";
	
			Args[0] = "/bin/tar";
			Args[1] = "cvfj";
			Args[2] = ListsFilename.c_str();

		
		int j=3;
		
		for (vector<string>::iterator I = PtrListFiles->begin(); I != PtrListFiles->end(); I++,j++)
			Args[j]=I->c_str();
		      
		
		Args[j] = 0;
		execvp(Args[0],(char **)Args);
		exit(100);
	}

	// Wait for tar to finish
	if (ExecWait(Process,"/bin/tar",false) == false)
		return _error->Error("tarbal of lists failed, perhaps the disk is full.");
	*/

	
	if(chdir(StartDir.c_str()) != 0)
		return false;
	closedir(D);
	return true;

}
/*}}}*/



//  GlobalDudf::AddMsgError - Append to msg error	/*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::AddMsgError(string MsgErrorPart)
{
	
	DudfErrorMsg+= MsgErrorPart;
	
}
/*}}}*/

//  GlobalDudf::PackagesInfo - Add info about a package /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::PackagesInfo(string PkgName, string PkgVerRel,string PkgStatus,string PkdDepBrokenName, 
		   string PkgDepConstrain,string PkgDepConstVersion,string PkgDepCandidate)
{
	
	DudfPackageInfo PkgElement;
	
	PkgElement.Name=PkgName;	
	PkgElement.VerRel=PkgVerRel;	
	PkgElement.Status=PkgStatus;
	PkgElement.DepBrokenName=PkdDepBrokenName;
	PkgElement.DepConstrain=PkgDepConstrain;
	PkgElement.DepConstVersion=PkgDepConstVersion;
	PkgElement.DepCandidate=PkgDepCandidate;
	
	PtrPkgInfo->push_back(PkgElement);
	
}
/*}}}*/




